//************************************************************
//
// main.cc
//
/** @file main.cc
 *
 * @brief Sample project for BPNGClientLib (blue PiraT 2 Client Library)
 *
 * @note Link static runtime library (/MTd resp. /MT).
 */
//************************************************************

#include <sys/stat.h>
#include <cerrno>
#include <ctime>
#include <iostream>
#include <direct.h>
#include <fstream>
#include <sstream>


#include "BPNGDefines.h"
#include "IBPNGClient.h"
#include "IBPNGClientListener.h"

#include "BPNGLoggerDetector.hh"
#include "RdbEventList.hh"


using namespace std;

static void sampleFunctionDownload(BP2Device device);
static void sampleFunctionOnlineConversion(BP2Device device);
static void sampleFunctionOfflineConversion();
static void sampleFunctionOfflineTSLConversion();
static void sampleFunctionConfiguration(BP2Device device);
static void getDevicesFromAbstractTSL(const string &dataSetBasePath, vector<BP2Device> &devices);
static string getLocalDateString();

int main()
{
    // Get list of all currently available blue PiraT 2 devices
    BPNGLoggerDetector detector;
    vector<BP2Device> devices = detector.getLoggerList(0);

    // select the device you want to work with
    BP2Device device;
    bool found = false;
    for (size_t i = 0; i < devices.size(); ++i)
    {
        if (devices[i].ipAddress == "192.168.0.233")
        {
            device = devices[i];
            found = true;
            break;
        }
    }

    if (!found)
        return 1;

    /** activate one of the sample functions */

    //sampleFunctionDownload(device);
    //sampleFunctionOnlineConversion(device);
    sampleFunctionOfflineConversion();
    //sampleFunctionConfiguration(device);
    sampleFunctionOfflineTSLConversion();
}


// We want to download all traces since last startup to an offline data set
static void sampleFunctionDownload(BP2Device device)
{
    IBPNGClient *client = getBPNGClient();

    // connect logger
    BOOL ret = client->connectLogger(device.ipAddress.c_str());
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to connect logger. " << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    ret = client->initOnline();
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to init online." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    IRdbEventList *list = client->getEventList();
    RdbEventList eventList(list);

    if (eventList.size() == 0)
    {
        cout << "Empty event list" << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    uint64_t startupId = 0;
    uint64_t endId = -1; //max value for uint64 to include everything in the id range

    // search last startup
    for (int i = eventList.size() - 1; i >= 0; --i)
    {
        if (eventList[i].type == STARTUP)
        {
            startupId = eventList[i].uniqueID;
            break;
        }
    }

    DataSpan span;
    span.type = DST_IDSPAN;
    span.start = startupId;
    span.end = endId;

    // if you want to download several spans, put them in a vector
    vector<DataSpan> spanVec;
    spanVec.push_back(span);

    ret = _mkdir("..\\testoutdir");
    if (ret != 0 && errno != EEXIST)
    {
        cout << "Failed to create output directory" << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    ret = client->downloadDataSpans(spanVec.size(), &spanVec[0], "..\\testoutdir\\BP2_Offline.zip", 0);
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to dowload data." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // disconnect
    client->disconnectLogger();
    // free memory
    client->release();
}

// We want to convert all CAN traces from the logger
// around the last Marker to CANoe asc and BLF format.
static void sampleFunctionOnlineConversion(BP2Device device)
{
    IBPNGClient *client = getBPNGClient();

    // connect logger
    BOOL ret = client->connectLogger(device.ipAddress.c_str());
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to connect logger." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    ret = client->initOnline();
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to init online." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    IRdbEventList* list = client->getEventList();
    RdbEventList eventList(list);
    if (eventList.size() == 0)
    {
        cout << "Empty event list" << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    uint64_t markerTimeStamp = 0;
    // search last marker
    for (int i = eventList.size() - 1; i >= 0; --i)
    {
        if (eventList[i].type == MARKER)
        {
            markerTimeStamp = eventList[i].timeStamp;
            break;
        }
    }

    if (markerTimeStamp == 0)
    {
        cout << "No marker found." << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // Ensure the out directory exists
    ret = _mkdir("..\\testoutdir");
    if (ret != 0 && errno != EEXIST)
    {
        cout << "Failed to create output directory" << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // Get a conversion set
    IConversionSet* conversionSet = client->createNewConversionSet();

    // The time span has to be 60s before and 60s after the marker
    uint64_t startTime = markerTimeStamp - 60 * 1000000; // in usec
    uint64_t endTime = markerTimeStamp + 60 * 1000000; // in usec

    // If you want to convert more than one span,
    // call this function several times
    conversionSet->addTimeSpan(startTime, endTime);


    // CAN #1 and CAN #2 are supposed to be written to one asc output file each.
    // CAN #3 and CAN #4 are supposed to be written together in another asc file.
    // All other CAN channels are supposed to be written together in one BLF file.
    const IChannelList* channels = client->getLoggerChannels();
    for (int i = 0; i < channels->getSize(); ++i)
    {
        ChannelType type = channels->getChannel(i)->getType();
        if (type != CH_CAN)
            continue;

        // Note: channel indices are zero based
        int index =  channels->getChannel(i)->getIndex();
        if (index == 0 || index == 1)
        {
            // CAN #1 and #2 in separate files
            // -1 as fileId parameter creates a separate file for this channel
            conversionSet->addChannel(type, index, CANOE, -1);
        }
        else if (index == 2 || index == 3)
        {
            // CAN #3 and #4 in the same file.
            // fileId != -1 will write all channels with the same format and same
            // file Id to the same output file (if procurable in accordance with
            // the format specification.
            conversionSet->addChannel(type, index, CANOE, 10);
        }
        else
        {
            // All other CAN channels to one BLF file.
            conversionSet->addChannel(type, index, BLF, 20);
        }
    }

    ret = client->convertData(conversionSet, "..\\testoutdir");
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to convert data." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // disconnect
    client->disconnectLogger();
    // free memory
    client->release();
}

// We want to convert all CAN traces from an Offline data set
// around the last Marker to CANoe asc and BLF format.
static void sampleFunctionOfflineConversion()
{
    IBPNGClient *client = getBPNGClient();

    // We use the sample Offline Data Set that was downloaded
    // with sampleFunctionDownload().
    // Its up to you to ensure an existing file.
    BOOL ret = client->initOffline("..\\testoutdir\\BP2_Offline.zip");
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to init offline." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    IRdbEventList* list = client->getEventList();
    RdbEventList eventList(list);
    if (eventList.size() == 0)
    {
        cout << "Empty event list" << endl;
        client->release();
        return;
    }

    uint64_t markerTimeStamp = 0;
    // search last marker
    for (int i = eventList.size() - 1; i >= 0; --i)
    {
        if (eventList[i].type == MARKER)
        {
            markerTimeStamp = eventList[i].timeStamp;
            break;
        }
    }

    if (markerTimeStamp == 0)
    {
        cout << "No marker found." << endl;
        client->release();
        return;
    }

    // Ensure the out directory exists
    ret = _mkdir("..\\testoutdir");
    if (ret != 0 && errno != EEXIST)
    {
        cout << "Failed to create output directory" << endl;
        client->release();
        return;
    }

    // Get a conversion set
    IConversionSet* conversionSet = client->createNewConversionSet();

    // The time span has to be 60s before and 60s after the marker
    uint64_t startTime = markerTimeStamp - 60 * 1000000; // in usec
    uint64_t endTime = markerTimeStamp + 60 * 1000000; // in usec

    // If you want to convert more than one span,
    // call this function several times
    conversionSet->addTimeSpan(startTime, endTime);

    // CAN #1 and CAN #2 are supposed to be written to one asc output file each.
    // CAN #3 and CAN #4 are supposed to be written together in another asc file.
    // All other CAN channels are supposed to be written together in one BLF file.
    const IChannelList* channels = client->getLoggerChannels();
    for (int i = 0; i < channels->getSize(); ++i)
    {
        ChannelType type = channels->getChannel(i)->getType();
        if (type != CH_CAN)
            continue;

        // Note: channel indices are zero based
        int index = channels->getChannel(i)->getIndex();
        if (index == 0 || index == 1)
        {
            // CAN #1 and #2 in separate files
            // -1 as fileId parameter creates a separate file for this channel
            conversionSet->addChannel(type, index, CANOE, -1);
        }
        else if (index == 2 || index == 3)
        {
            // CAN #3 and #4 in the same file.
            // fileId != -1 will write all channels with the same format and same
            // file Id to the same output file (if procurable in accordance with
            // the format specification.
            conversionSet->addChannel(type, index, CANOE, 10);
        }
        else
        {
            // All other CAN channels to one BLF file.
            conversionSet->addChannel(type, index, BLF, 20);
        }
    }

    ret = client->convertData(conversionSet, "..\\testoutdir");
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to convert data." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    // free memory
    client->release();
}

// We want to convert all A/I, D/I traces from an Offline TSL data set
// to XTMT
static void sampleFunctionOfflineTSLConversion()
{
    // informations about tsl can be read out of 'abstract_TSL.txt' in root of offline data set
    string offlineDataSetPath = "C:\\Users\\qi10421\\Desktop\\Offline_Angelsachsen_20160111_110108_20160113_152844";
    vector<BP2Device> devices;
    getDevicesFromAbstractTSL(offlineDataSetPath, devices);

    // create a tsl client: parameter is number of tsl member
    IBPNGClient *client = getTSLClient(devices.size());

    // Its up to you to ensure an existing file.
    BOOL ret = client->initOffline(offlineDataSetPath.c_str());
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to init offline." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    // Ensure the out directory exists
    ret = _mkdir("..\\testoutdir");
    if (ret != 0 && errno != EEXIST)
    {
        cout << "Failed to create output directory" << endl;
        client->release();
        return;
    }

    // Get a conversion set
    IConversionSet* conversionSet = client->createNewConversionSet();

    //select all spans
    conversionSet->addTimeSpan(0, 0xFFFFFFFFFFFFFFFFUL);

    // All A/I, D/I channels are supposed to be written in separated XTMT files.
    const IChannelList* channels = client->getLoggerChannels();
    for (int i = 0; i < channels->getSize(); ++i)
    {
        ChannelType type = channels->getChannel(i)->getType();
        if (type == CH_DIGITAL_IN || type == CH_ANALOG_IN)
        {
            // -1 as fileId parameter creates a separate file for this channel
            // for tsl the offset and mainboardnumber fields are needed
            conversionSet->addChannel(type, channels->getChannel(i)->getIndex(), XTMT, -1, channels->getChannel(i)->getOffset(), channels->getChannel(i)->getMainboardNumber());
        }
    }

    ret = client->convertData(conversionSet, "..\\testoutdir");
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to convert data." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    // free memory
    client->release();
}



// This function shows how to:
//   - download the configuration
//   - reconfigurate the logger device
//   - set the default config
static void sampleFunctionConfiguration(BP2Device device)
{
    IBPNGClient *client = getBPNGClient();

    BOOL ret = client->connectLogger(device.ipAddress.c_str());
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to connect logger." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->release();
        return;
    }

    // save current config
    string targetPath = "..\\testoutdir\\bpng_[" + device.mainboardNr + "]_[" + getLocalDateString() + "].zip";
    ret = client->getConfig(targetPath.c_str());
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to download configuration." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // here you could change the downloaded configuration
    // by extracting it and modifying the xml files
    // see documentation of IBPNGClient::getConfig()
    // or IBPNGClient::reconfigLogger().
    // the new config archive needs a date in its filename in this form: YYYY-MM-DD_HH-MM-SS

    // We use the same config that we downloaded
    string newConfigPath = targetPath;
    ret = client->reconfigLogger(newConfigPath.c_str());
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to reconfigurate the logger." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // Setting the default config
    ret = client->setDefaultConfig();
    if (ret == 0)
    {
        BPNGError err = client->getLastError();
        cout << "Failed to set default config to the logger." << endl;
        cout << "BPNGErrCode: " << err.code << ", " << err.msg << endl;
        client->disconnectLogger();
        client->release();
        return;
    }

    // disconnect
    client->disconnectLogger();
    // free memory
    client->release();
}

static void getDevicesFromAbstractTSL(const string &dataSetBasePath, vector<BP2Device> &devices)
{
	string source = dataSetBasePath + "//abstract_TSL.txt";
    ifstream file(source.c_str());
    string str;
    while (getline(file, str))
    {
        int equalPos = str.find("=");
        if (equalPos == string::npos)
        {
            continue;
        }

        if (str.find("device_ips") != string::npos)
        {
            stringstream lookup(str.substr(equalPos + 1, str.length() - equalPos - 1));
            while (getline(lookup, str, ';'))
            {
                BP2Device dev;
                dev.ipAddress = str;
                devices.push_back(dev);
            }
        }

        if (str.find("device_names") != string::npos)
        {
            stringstream lookup(str.substr(equalPos + 1, str.length() - equalPos - 1));
            int i = 0;
            while (getline(lookup, str, ';'))
            {
                if (i >= devices.size())
                {
                    break;
                }
                devices.at(i).name = str;
                ++i;
            }
        }

        if (str.find("device_mbnrs") != string::npos)
        {
            stringstream lookup(str.substr(equalPos + 1, str.length() - equalPos - 1));
            int i = 0;
            while (getline(lookup, str, ';'))
            {
                if (i >= devices.size())
                {
                    break;
                }
                devices.at(i).mainboardNr = str;
                ++i;
            }
        }
    }
}

static string getLocalDateString()
{
    time_t timeObj;
    time(&timeObj);
    tm *pTime = localtime(&timeObj);
    char buffer[100];
    int hour = pTime->tm_isdst ? pTime->tm_hour + 1 : pTime->tm_hour;
    sprintf(buffer, "%d-%02d-%02d_%02d-%02d-%02d", pTime->tm_year + 1900, pTime->tm_mon + 1, pTime->tm_mday, hour, pTime->tm_min, pTime->tm_sec);

    return buffer;
}
